# RethinkDB wrappers
# Copyright (C) 2020-2021  Nguyễn Gia Phong
#
# This file is part of Acanban.
#
# Acanban is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as published
# by the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Acanban is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with Acanban.  If not, see <https://www.gnu.org/licenses/>.

from typing import Any, Awaitable, Tuple

from quart import current_app
from rethinkdb import r

__all__ = ['RethinkObject']


async def get(table: str, key: str, field: str) -> Any:
    """Return DB[table][key][field]."""
    async with current_app.db_pool.connection() as conn:
        return await r.table(table).get(key)[field].run(conn)


class RethinkObject:
    """RethinkDB object wrapper mixin.

    Exceptions raised by RethinkDB are expected to be handled by caller.
    """

    slots: Tuple[str, ...]
    table: str
    key: str

    def __getattr__(self, name: str) -> Awaitable:
        if name not in self.slots:
            cls = self.__class__.__name__
            raise AttributeError(f'{cls!r} object has no attribute {name!r}')
        return get(self.table, self.key, name)

    def __setattr__(self, name: str, value: Any) -> None:
        if name in self.slots:
            cls = self.__class__.__name__
            raise AttributeError(
                f'{cls}.update must be used for dynamic attribute {name!r}')
        super().__setattr__(name, value)

    async def update(self, **kwargs: Any) -> None:
        """Update the RethinkDB object from given keys and values.

        This is defined instead of `__setattr__` because
        it is impossible to await an assignment statement.

        Raise
        -----
        ValueError
            If one or more field name in kwargs undeclared in slots.
        """
        invalid = ', '.join(repr(field) for field in kwargs
                            if field not in self.slots)
        if invalid:
            cls = self.__class__.__name__
            raise ValueError(f'cannot dynamically set {invalid} in {cls!r}')
        async with current_app.db_pool.connection() as conn:
            await r.table(self.table).get(self.key).update(kwargs).run(conn)



# IPFS proxy
# Copyright (C) 2021  Nguyễn Gia Phong
#
# This file is part of Acanban.
#
# Acanban is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as published
# by the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Acanban is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with Acanban.  If not, see <https://www.gnu.org/licenses/>.

from typing import AsyncIterator
from urllib.parse import urlsplit, urlunsplit

from cid import CIDv1, make_cid
from httpx import Response
from quart import (Blueprint, ResponseReturnValue,
                   current_app, redirect, request)
from quart.exceptions import NotFound, RequestTimeout
from quart_auth import current_user
from rethinkdb import r
from trio import TooSlowError, fail_after

__all__ = ['add', 'blueprint']

blueprint = Blueprint('ipfs', __name__)


def ensure_cidv1(multihash: str) -> str:
    """Return the given multihash encoded as base58 in CIDv1."""
    cid = make_cid(multihash)
    return (cid if isinstance(cid, CIDv1) else cid.to_v1()).encode().decode()


async def add() -> str:
    """Proxy file upload to IPFS add and return UUID in table files."""
    headers = {'Content-Length': request.headers['Content-Length'],
               'Content-Type': request.headers['Content-Type']}
    try:
        with fail_after(current_app.config['BODY_TIMEOUT']):
            response = await current_app.ipfs_api.post(
                '/add', headers=headers, data=request.body)
    except TooSlowError:  # pragma: no cover
        raise RequestTimeout
    response.raise_for_status()

    file = response.json()
    async with current_app.db_pool.connection() as conn:
        result = await r.table('files').insert(dict(
            cid=ensure_cidv1(file['Hash']),
            name=file['Name'], size=int(file['Size']),
            time=r.now(), user=current_user.key)).run(conn)
    return result['generated_keys'][0]


def fallback() -> ResponseReturnValue:
    """Redirect the current request to the fallback IPFS gateway."""
    base = urlsplit(current_app.ipfs_config['gateway']['fallback'])
    url = urlsplit(request.url)
    elements = base.scheme, base.netloc, url.path, url.query, url.fragment
    return redirect(urlunsplit(elements), 301)


@blueprint.route('/ipns/<path:path>')
async def redirect_ipns(path: str) -> ResponseReturnValue:
    """Redirect IPNS URLs to the fallback gateway."""
    return fallback()


async def stream(response: Response) -> AsyncIterator[bytes]:
    """Return the response's raw stream."""
    try:
        async for chunk in response.aiter_raw(): yield chunk
    finally:
        await response.aclose()


@blueprint.route('/ipfs/<path:path>')
async def proxy_gateway(path: str) -> ResponseReturnValue:
    """Proxy IPFS gateway.

    If the file was not added through Acanban,
    redirect to the fallback gateway instead.
    """
    try:
        cid = ensure_cidv1(path.split('/', 1)[0])
    except ValueError:
        raise NotFound
    count = r.table('files').get_all(cid, index='cid').count()
    async with current_app.db_pool.connection() as connection:
        if not await count.run(connection): return fallback()

    client = current_app.ipfs_gateway
    proxied_request = client.build_request(
        'GET', path, params=request.args,
        headers={**request.headers, 'Host': client.base_url.netloc})
    response = await client.send(proxied_request, stream=True)
    return stream(response), response.status_code, response.headers.items()



# Configuration parser
# Copyright (C) 2020-2021  Nguyễn Gia Phong
#
# This file is part of Acanban.
#
# Acanban is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as published
# by the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Acanban is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with Acanban.  If not, see <https://www.gnu.org/licenses/>.

from os.path import isfile, join
from typing import Any, MutableMapping, Sequence

import toml
from appdirs import site_config_dir, user_config_dir
from hypercorn.config import Config as HyperConf

TomMapping = MutableMapping[str, Any]

CONFIG_DIRS = user_config_dir('acanban'), site_config_dir('acanban')
ACANBAN_DEFAULT: TomMapping = {}
IPFS_DEFAULT: TomMapping = {
    'api': {'base': 'http://127.0.0.1:5001/api/v0'},
    'gateway': {'base': 'http://127.0.0.1:8080/ipfs',
                'fallback': 'https://ipfs.io'}}
RETHINKDB_DEFAULT: TomMapping = {'db': 'test'}


def acanban_config(dirs: Sequence[str] = CONFIG_DIRS) -> TomMapping:
    """Return Acanban configuration first found in given directories."""
    for directory in dirs:
        file = join(directory, 'acanban.toml')
        if isfile(file): return toml.load(file)
    return ACANBAN_DEFAULT


def ipfs_config(dirs: Sequence[str] = CONFIG_DIRS) -> TomMapping:
    """Return IPFS configuration first found in given directories."""
    for directory in dirs:
        file = join(directory, 'ipfs.toml')
        if isfile(file): return toml.load(file)
    return IPFS_DEFAULT


def hypercorn_config(dirs: Sequence[str] = CONFIG_DIRS) -> HyperConf:
    """Return Hypercorn configuration first found in given directories."""
    for directory in dirs:
        file = join(directory, 'hypercorn.toml')
        if isfile(file): return HyperConf.from_toml(file)
    return HyperConf()


def rethinkdb_config(dirs: Sequence[str] = CONFIG_DIRS) -> TomMapping:
    """Return RethinkDB configuration first found in given directories."""
    for directory in dirs:
        file = join(directory, 'rethinkdb.toml')
        if isfile(file): return toml.load(file)
    return RETHINKDB_DEFAULT
